/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/**@file
 *
 * @defgroup ble_sdk_app_bootloader_main main.c
 * @{
 * @ingroup dfu_bootloader_api
 * @brief Bootloader project main file.
 *
 * -# Receive start data packet. 
 * -# Based on start packet, prepare NVM area to store received data. 
 * -# Receive data packet. 
 * -# Validate data packet.
 * -# Write Data packet to NVM.
 * -# If not finished - Wait for next packet.
 * -# Receive stop data packet.
 * -# Activate Image, boot application.
 *
 */
#include "dfu_transport.h"
#include "bootloader.h"
#include "bootloader_util.h"
#include <stdint.h>
#include <string.h>
#include <stddef.h>
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_soc.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "ble.h"
#include "nrf.h"
#include "ble_hci.h"
#include "app_scheduler.h"
#include "app_timer_appsh.h"
#include "nrf_error.h"
#include "bsp.h"
#include "softdevice_handler_appsh.h"
#include "pstorage_platform.h"
#include "nrf_mbr.h"

#if BUTTONS_NUMBER < 1
#error "Not enough buttons on board"
#endif

#if LEDS_NUMBER < 1
#error "Not enough LEDs on board"
#endif

#define IS_SRVC_CHANGED_CHARACT_PRESENT 1                                                       /**< Include the service_changed characteristic. For DFU this should normally be the case. */

#define BOOTLOADER_BUTTON               BSP_BUTTON_3                                            /**< Button used to enter SW update mode. */
#define UPDATE_IN_PROGRESS_LED          BSP_LED_2                                               /**< Led used to indicate that DFU is active. */

#define APP_TIMER_PRESCALER             0                                                       /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE         4                                                       /**< Size of timer operation queues. */

#define SCHED_MAX_EVENT_DATA_SIZE       MAX(APP_TIMER_SCHED_EVT_SIZE, 0)                        /**< Maximum size of scheduler events. */

#define SCHED_QUEUE_SIZE                20                                                      /**< Maximum number of events in the scheduler queue. */


/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze 
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num    Line number of the failing ASSERT call.
 * @param[in] file_name   File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(0xDEADBEEF, line_num, p_file_name);
}


/**@brief Function for initialization of LEDs.
 */
static void leds_init(void)
{   
    // [hv] Configure GPIO pin from LED_START to LED_STOP as a normal output
    nrf_gpio_range_cfg_output(LED_START, LED_STOP);
    
    // [hv] Setting LED pin to high. LEDs are active low, so this turn them off
    nrf_gpio_pins_set(LEDS_MASK);
}


/**@brief Function for initializing the timer handler module (app_timer).
 */
static void timers_init(void)
{
    // Initialize timer module, making it use the scheduler.
    APP_TIMER_APPSH_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, true);
}


/**@brief Function for initializing the button module.
 */
static void buttons_init(void)
{   
    // [hv] Configure GPIO pin used for Button input as an input, preset with a high value
    // and "sense"/wake up when the pin is pull low...?
    nrf_gpio_cfg_sense_input(BOOTLOADER_BUTTON,
                             BUTTON_PULL, 
                             NRF_GPIO_PIN_SENSE_LOW);

}


/**@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack
 *          event has been received.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 */
static void sys_evt_dispatch(uint32_t event)
{
    pstorage_sys_event_handler(event);
}


/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 *
 * @param[in] init_softdevice  true if SoftDevice should be initialized. The SoftDevice must only 
 *                             be initialized if a chip reset has occured. Soft reset from 
 *                             application must not reinitialize the SoftDevice.
 */
static void ble_stack_init(bool init_softdevice)
{
    uint32_t         err_code;
    sd_mbr_command_t com = {SD_MBR_COMMAND_INIT_SD, };

    if (init_softdevice)
    {
        err_code = sd_mbr_command(&com);
        APP_ERROR_CHECK(err_code);
    }
    
    err_code = sd_softdevice_vector_table_base_set(BOOTLOADER_REGION_START);
    APP_ERROR_CHECK(err_code);
   
    SOFTDEVICE_HANDLER_APPSH_INIT(NRF_CLOCK_LFCLKSRC_RC_250_PPM_4000MS_CALIBRATION, true);

    // Enable BLE stack 
    ble_enable_params_t ble_enable_params;
    memset(&ble_enable_params, 0, sizeof(ble_enable_params));
    
    // Below code line is needed for s130. For s110 is inrrelevant - but executable
    // can run with both s130 and s110.
    ble_enable_params.gatts_enable_params.attr_tab_size   = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;

    ble_enable_params.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
    err_code = sd_ble_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);
    
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for event scheduler initialization.
 */
static void scheduler_init(void)
{
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);
}


/**@brief Function for bootloader main entry.
 */
int main(void)
{
    uint32_t err_code;
    bool     dfu_start = false;
  
    // [hv] Check Power Periph's Retention RAM if it is storing value signifying 
    // that device should be boooted into Bootloader mode. Probably this value 
    // can be set by the DFU Service
    bool     app_reset = (NRF_POWER->GPREGRET == BOOTLOADER_DFU_START);
    
    if (app_reset)
    {
        // [hv] Clearing the Power Retention RAM
        NRF_POWER->GPREGRET = 0;
    }
    
    // [hv] Set GPIO pins driving the LEDs as output and pull them high, 
    // which turn off the LEDs
    // leds_init(); // [hv] comment out because not using PCA10028

    // This check ensures that the defined fields in the bootloader corresponds with actual
    // setting in the nRF51 chip.
    APP_ERROR_CHECK_BOOL(*((uint32_t *)NRF_UICR_BOOT_START_ADDRESS) == BOOTLOADER_REGION_START);
    APP_ERROR_CHECK_BOOL(NRF_FICR->CODEPAGESIZE == CODE_PAGE_SIZE);

    // Initialize.
    
    // [hv] initiate timers for purpose which remains unclear
    timers_init();
    
    
    // [hv] Configure GPIO pin used for Button input as an input, preset with a high value
    // and "sense"/wake up when the pin is pull low...?
    // buttons_init(); // [hv] comment out because not using PCA10028
    
    // [hv] Seems nothing board specific in this function
    (void)bootloader_init();
    
    // [hv] Check if a SoftDevice update is in progress
    if (bootloader_dfu_sd_in_progress())
    {
        // [hv] turn on LED (active low), so need to be removed
        // nrf_gpio_pin_clear(UPDATE_IN_PROGRESS_LED); // [hv] comment out because not using PCA10028

        err_code = bootloader_dfu_sd_update_continue();
        APP_ERROR_CHECK(err_code);
        
        // [hv] param: true if SoftDevice should be initialized. The SoftDevice must only 
        // be initialized if a chip reset has occured. Soft reset from 
        // application must not reinitialize the SoftDevice.
        // So, in other word, if the app soft reset and write onto the Power Periph's 
        // Retention Reg a BOOTLOADER_DFU_START, then SoftDevice should not be reinitialize
        ble_stack_init(!app_reset);
        
        // [hv] Seems nothing board specific here
        scheduler_init();
        
        // [hv] Seems nothing board specific here
        err_code = bootloader_dfu_sd_update_finalize();
        APP_ERROR_CHECK(err_code);
        
        // [hv] turn off LED (active low), so need to be removed
        // nrf_gpio_pin_set(UPDATE_IN_PROGRESS_LED); // [hv] comment out because not using PCA10028
    }
    else
    {
        // If stack is present then continue initialization of bootloader.
        ble_stack_init(!app_reset);
        scheduler_init();
    }
    
    
    dfu_start  = app_reset;
    
    // [hv] buttons are Normally High so if it is 0 it is pressed
    // BOOTLOADER_BUTTON is BSP_BUTTON_3 which actually is BUTTON_4 on the PCA10028 board...
    // So, dfu_start |= (true if button pressed, false otherwise)
    // So for the nrf51822 chip this line could probably be removed
    // dfu_start |= ((nrf_gpio_pin_read(BOOTLOADER_BUTTON) == 0) ? true: false); // [hv] comment out because not using PCA10028
    
    // [hv] Check if there is a valid Application in bank 0 ? or if dfu_start is true
    if (dfu_start || (!bootloader_app_is_valid(DFU_BANK_0_REGION_START)))
    {   
        // [hv] turn on LED (active low), so need to be removed
        //nrf_gpio_pin_clear(UPDATE_IN_PROGRESS_LED); // [hv] comment out because not using PCA10028

        // Initiate an update of the firmware.
        err_code = bootloader_dfu_start();
        APP_ERROR_CHECK(err_code);
        
        // [hv] turn off LED (active low), so need to be removed
        //nrf_gpio_pin_set(UPDATE_IN_PROGRESS_LED); // [hv] comment out because not using PCA10028
    }
    
    // [hv] Check if app is valid AND no SoftDevice update in progress.
    // if true starts app
    if (bootloader_app_is_valid(DFU_BANK_0_REGION_START) && !bootloader_dfu_sd_in_progress())
    {
        // Select a bank region to use as application region.
        // @note: Only applications running from DFU_BANK_0_REGION_START is supported.
        bootloader_app_start(DFU_BANK_0_REGION_START);
    }
    
    // [hv] reset system
    NVIC_SystemReset();
}
